/*************************************************************************
 *
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * Copyright 2008 by Nakata, Maho
 *
 * MPACK - multiple precision arithmetic library
 *
 * This file is part of MPACK.
 *
 * MPACK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * OpenOffice.org is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details
 * (a copy is included in the LICENSE file that accompanied this code).
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with OpenOffice.org.  If not, see
 * <http://www.openoffice.org/license.html>
 * for a copy of the LGPLv3 License.
 *
 ************************************************************************/
/*
Copyright (c) 1992-2007 The University of Tennessee.  All rights reserved.

$COPYRIGHT$

Additional copyrights may follow

$HEADER$

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer. 
  
- Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer listed
  in this license in the documentation and/or other materials
  provided with the distribution.
  
- Neither the name of the copyright holders nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.
  
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT  
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT  
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/

/*
http://www.netlib.org/blas/zhemm.f
Chemm performs one of the matrix-matrix operations
 C := alpha*A*B + beta*C,
or
 C := alpha*B*A + beta*C,
where alpha and beta are scalars, A is an hermitian matrix and B and
C are m by n matrices.
*/

#include <mblas.h>

void
Chemm(const char *side, const char *uplo, int m, int n, mpc_class alpha,
    mpc_class * A, int lda, mpc_class * B, int ldb, mpc_class beta,
    mpc_class * C, int ldc)
{
    int nrowa, upper, info;
    mpc_class Zero, One;

    Zero.re = 0.0;
    Zero.im = 0.0;
    One.re = 1.0;
    One.im = 0.0;
    mpc_class temp1, temp2;

    //set nrowa as the number of rows of a.
    if (Mlsame(side, "L"))
	nrowa = m;
    else
	nrowa = n;
    upper = Mlsame(uplo, "U");

    //test the input parameters.
    info = 0;
    if ((!Mlsame(side, "L")) && (!Mlsame(side, "R")))
	info = 1;
    else if ((!upper) && (!Mlsame(uplo, "L")))
	info = 2;
    else if (m < 0)
	info = 3;
    else if (n < 0)
	info = 4;
    else if (lda < max(1, nrowa))
	info = 7;
    else if (ldb < max(1, m))
	info = 9;
    else if (ldc < max(1, m))
	info = 12;
    if (info != 0) {
	Mxerbla("Chemm ", info);
	return;
    }
    //quick return if possible.
    if ((m == 0) || (n == 0) || ((alpha == Zero) && (beta == One)))
	return;

    //and when alpha==Zero.
    if (alpha == Zero) {
	if (beta == Zero) {
	    for (int j = 0; j < n; j++) {
		for (int i = 0; i < m; i++) {
		    C[i + j * ldc] = Zero;
		}
	    }
	} else {
	    for (int j = 0; j < n; j++) {
		for (int i = 0; i < m; i++) {
		    C[i + j * ldc] = beta * C[i + j * ldc];
		}
	    }
	}
	return;
    }
    //start the operations.
    if (Mlsame(side, "L")) {
	//Form C := alpha*A*B + beta*C.
	if (upper) {
	    for (int j = 0; j < n; j++) {
		for (int i = 0; i < m; i++) {
		    temp1 = alpha * B[i + j * ldb];
		    temp2 = Zero;
		    for (int k = 0; k < i; k++) {
			C[k + j * ldc] =
			    C[k + j * ldc] + temp1 * A[k + i * lda];
			temp2 = temp2 + B[k + j * ldb] * conjg(A[k + i * lda]);
		    }
		    if (beta == Zero) {
			C[i + j * ldc] =
			    temp1 * real(A[i + i * lda]) + alpha * temp2;
		    } else {
			C[i + j * ldc] =
			    beta * C[i + j * ldc] + temp1 * real(A[i +
				i * lda]) + alpha * temp2;
		    }
		}
	    }
	} else {
	    //Form  C := alpha*A*B + beta*C.
	    for (int j = 0; j < n; j++) {
		for (int i = m - 1; i >= 0; i--) {
		    temp1 = alpha * B[i + j * ldb];
		    temp2 = Zero;
		    for (int k = i + 1; k < m; k++) {
			C[k + j * ldc] =
			    C[k + j * ldc] + temp1 * A[k + i * lda];
			temp2 = temp2 + B[k + j * ldb] * conjg(A[k + i * lda]);
		    }
		    if (beta == Zero)
			C[i + j * ldc] =
			    temp1 * real(A[i + i * lda]) + alpha * temp2;
		    else
			C[i + j * ldc] =
			    beta * C[i + j * ldc] + temp1 * real(A[i +
				i * lda]) + alpha * temp2;

		}
	    }
	}
    } else {
	//Form  C := alpha*B*A + beta*C.
	for (int j = 0; j < n; j++) {
	    temp1 = alpha * real(A[j + j * lda]);
	    if (beta == Zero) {
		for (int i = 0; i < m; i++) {
		    C[i + j * ldc] = temp1 * B[i + j * ldb];
		}
	    } else {
		for (int i = 0; i < m; i++) {
		    C[i + j * ldc] =
			beta * C[i + j * ldc] + temp1 * B[i + j * ldb];
		}
	    }
	    for (int k = 0; k < j; k++) {
		if (upper)
		    temp1 = alpha * A[k + j * lda];
		else
		    temp1 = alpha * conjg(A[j + k * lda]);

		for (int i = 0; i < m; i++) {
		    C[i + j * ldc] = C[i + j * ldc] + temp1 * B[i + k * ldb];
		}
	    }
	    for (int k = j + 1; k < n; k++) {
		if (upper)
		    temp1 = alpha * conjg(A[j + k * lda]);
		else
		    temp1 = alpha * A[k + j * lda];

		for (int i = 0; i < m; i++) {
		    C[i + j * ldc] = C[i + j * ldc] + temp1 * B[i + k * ldb];
		}
	    }
	}
    }
    return;
}
